<?php

namespace Iwaydev\Localization\Services;

use Iwaydev\Localization\LocalizationClient;
use JsonException;
use Psr\Log\LoggerInterface;
use Psr\SimpleCache\CacheInterface;
use Psr\SimpleCache\InvalidArgumentException;
use Throwable;

/**
 * Class LocalizationService
 *
 * @package App\Services
 */
class LocalizationService
{
    private CacheInterface $cache;
    private int $cacheLifetime;
    private LocalizationClient $client;
    private LoggerInterface $logger;
    private string $projectSlug;
    private string $password;
    private string $token;

    /**
     * LocalizationService constructor.
     *
     * @param LocalizationClient $client
     * @param string $projectSlug
     * @param string $password
     * @param LoggerInterface $logger
     * @param CacheInterface $cache
     * @param int $cacheLifetime
     *
     * @throws JsonException
     * @throws Throwable
     */
    public function __construct(
        LocalizationClient $client,
        string $projectSlug,
        string $password,
        LoggerInterface $logger,
        CacheInterface $cache,
        int $cacheLifetime
    ) {
        $this->client = $client;
        $this->projectSlug = $projectSlug;
        $this->password = $password;

        $this->setLogger($logger);
        $this->cache = $cache;
        $this->cacheLifetime = $cacheLifetime;

        $this->login();
    }

    /**
     * Установить логгер
     *
     * @param LoggerInterface $logger
     *
     * @return $this
     */
    public function setLogger(LoggerInterface $logger): self
    {
        $this->logger = $logger;

        return $this;
    }

    /**
     * Сделать запись уровня "info" в лог, если логгер установлен
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logInfo(string $message, array $context = []): void
    {
        if (!empty($this->logger)) {
            $this->logger->info($this->getXUuid() . ' ' . $message, $context);
        }
    }

    /**
     * Сделать запись уровня "error" в лог, если логгер установлен
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logError(string $message, array $context = []): void
    {
        if (!empty($this->logger)) {
            $this->logger->error($this->getXUuid() . ' ' . $message, $context);
        }
    }

    /**
     * Логируем исключение
     *
     * @param Throwable $throwable Экзепляр исключения
     */
    private function logException(Throwable $throwable): void
    {
        $this->logError('Исключение!');
        $this->logError('Сообщение: ' . $throwable->getMessage());
        $this->logError('Код: ' . $throwable->getCode());
        $this->logError('Файл: ' . $throwable->getFile() . '(' . $throwable->getLine() . ')');
        $this->logError('Трейс: ' . PHP_EOL . $throwable->getTraceAsString());
    }

    /**
     * Обновляем переводы
     *
     * @throws InvalidArgumentException
     * @throws JsonException
     * @throws Throwable
     */
    public function updateLocalization(): void
    {
        try {
            $this->logInfo('Обновление начинается');

            $response = $this->checkStatusProject();
            if ($response['status'] === false) {
                $this->logInfo('.Запрашиваю метки с переводами для ' . $this->projectSlug);
                $localizations = $this->getAllLocalizations();
                $this->logInfo('..Получено ' . count($localizations) . ' меток');

                $this->logInfo('.Сохраняю метки в кэш');
                $this->saveLocalizationToCash($this->prepareLocalizationToSave($localizations));
                $this->logInfo('..Метки сохранены в кэше');

                $this->setHash($response['hash']);
                $this->logInfo('.Обновил хэш статуса проекта');
            } else {
                $this->logInfo('.Переводы еще актуальны');
            }

            $this->logInfo('Обновление завершено');
        } catch (Throwable $throwable) {
            $this->logException($throwable);
            throw $throwable;
        }
    }

    /**
     * Подготавливает метки с переводами к сохранению
     *
     * @param array $localizations
     *
     * @return array
     */
    private function prepareLocalizationToSave(array $localizations): array
    {
        $blackListAttributes = ['id', 'title', 'slug', 'project_id', 'created_at', 'updated_at'];

        return array_map(function ($item) use ($blackListAttributes) {
            return array_filter(
                $item,
                function ($value, $key) use ($blackListAttributes) {
                    return !in_array($key, $blackListAttributes) && !empty($value);
                },
                ARRAY_FILTER_USE_BOTH
            );
        }, $localizations);
    }

    /**
     * Сохраняем метки переводов в кэше
     *
     * @param array $localizations
     *
     * @throws InvalidArgumentException
     */
    private function saveLocalizationToCash(array $localizations): void
    {
        if (!empty($localizations)) {
            $this->cache->setMultiple($localizations, $this->cacheLifetime);
        }
    }

    /**
     * Возвращает все переводы по проекту
     *
     * @param int $page
     * @param array $items
     *
     * @return array
     * @throws JsonException
     */
    public function getAllLocalizations(int $page = 1, array $items = []): array
    {
        $response = $this->client->getLocalizations($this->projectSlug, $page);
        $items = array_merge($items, $response['items']);
        if ($page < $response['page_count']) {
            $page++;
            $items = $this->getAllLocalizations($page, $items);
        }

        return $items;
    }

    /**
     * Установить хэш
     *
     * @param string $hash
     *
     * @throws InvalidArgumentException
     */
    private function setHash(string $hash): void
    {
        $this->cache->set($this->getHashKey(), $hash, $this->cacheLifetime);
    }

    /**
     * Возвращает текущий хэш
     *
     * @return string
     * @throws InvalidArgumentException
     */
    private function getHash(): string
    {
        return $this->cache->get($this->getHashKey(), '');
    }

    /**
     * Возвращает ключ хэша для кэша
     *
     * @return string
     */
    private function getHashKey(): string
    {
        return 'hash_' . $this->projectSlug;
    }

    /**
     * Авторизация
     *
     * @return $this
     * @throws JsonException
     * @throws Throwable
     */
    public function login(): self
    {
        try {
            $this->token = $this->client->caching()->login($this->projectSlug, $this->password);

            return $this;
        } catch (Throwable $throwable) {
            $this->logException($throwable);
            throw $throwable;
        }
    }

    /**
     * Нужно ли обновлять переводы
     *
     * @return array
     * @throws InvalidArgumentException
     * @throws JsonException
     */
    public function checkStatusProject(): array
    {
        return $this->client->auth($this->token)->checkStatusProject($this->projectSlug, $this->getHash());
    }

    /**
     * Возвращает текущий X-UUID
     *
     * @return string
     */
    public function getXUuid(): string
    {
        return $this->client->getXUuid();
    }

    /**
     * Перевести метку
     *
     * @param string $slug
     * @param string $lang
     * @param array $parameters
     *
     * @return string
     * @throws InvalidArgumentException
     */
    public static function t(string $slug, string $lang, array $parameters = []): string
    {
        $localizations = self::getLocalizationBySlug($slug);
        if (empty($localizations) || !isset($localizations[$lang])) {
            return $slug;
        }

        $localization = $localizations[$lang];
        foreach ($parameters as $key => $value) {
            $localization = str_replace('{' . $key . '}', $value, $localization);
        }

        return $localization;
    }

    /**
     * Получить переводы для указанной метки
     *
     * @param string $slug
     *
     * @return array
     * @throws InvalidArgumentException
     */
    private static function getLocalizationBySlug(string $slug): array
    {
        return cache()->store()->get($slug, []);
    }
}
